""" plot massflow and enthalpy of generators by groups.

Massflow will be stacked within a group.  An average enthalpy will be included
for each group.

Expects 'settings.json' file whereever this script is launched.  It should look
something like this:

{
    "external_scenario_names": {
        "s3A"   : "s3A"   ,
        "s3B"   : "s3B"   ,
        "s4A"   : "s4A"   ,
        "s4A1"  : "s4A1"  ,
        "s4A2"  : "s4A2"  ,
        "s4C"   : "s4C"
    },
    "use_spec_json": true,
    "dir_to_extract" : [
        "../s3A-V78",
        "../s3A2-V78",
        "../s3Ab-V78",
        "../s3Ac-V78"
    ],
    "production": {
        "block": ".....",
        "gener": "(MT|MW|ET|EW)..."
    },
    "injection": {
        "block": ".....",
        "gener": "(I[ABCD]|HI)..."
    }
}

"""

from t2data import *

from gimfu.t2listingh5 import t2listingh5

from gimfu.scenario_extraction import external_name
from gimfu.scenario_extraction import average_enthalpy
from gimfu.scenario_extraction import matching_geners_from_lsts
from gimfu.scenario_extraction import to_year, to_kjkg, to_thr_rev, to_ktd_rev
from gimfu.scenario_extraction import assign_kgs, assign_jkg, assign_sec
from gimfu.multiple_listings_plotting import get_gener_history
from gimfu.scenario_spec_report import collate_well_stacks
from gimfu.easy_date import year_fraction_to_date_str

import json
import csv
import glob
import re
import os
import os.path
import pickle
from pprint import pprint as pp
from functools import partial

import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
from cycler import cycler

import numpy as np

# plt.style.use("seaborn-v0_8-dark")
mpl.rcParams['axes.formatter.useoffset'] = False
colors = np.vstack([
    plt.cm.Set1.colors,
    plt.cm.Set2.colors,
    plt.cm.Set3.colors,
    # plt.cm.Pastel1.colors,
    # plt.cm.Pastel2.colors,
    plt.cm.tab20b.colors,
    plt.cm.tab20.colors,
    # plt.cm.tab20c.colors
])
mpl.rcParams["axes.prop_cycle"] = cycler(color=colors)


def wellstacks_from_spec(spec_json):
    """ return a dict of {"well_stack": ["Well1", "Well2"]} """
    stacks, wells = collate_well_stacks(spec_json)
    wellstacks = {}
    for stackname, spec in stacks.items():
        wellstacks[stackname] = spec["wells"]
    return wellstacks

def grouping_from_spec(spec_json):
    """ returns a grouping dict based on the spec.

        Basically reverse of the well_geners from collate_well_stacks().
        grouping is a dict {(block, gener): 'GRP NAME', ...}
    """
    stacks, wells = collate_well_stacks(spec_json)
    grouping = {}
    for well, bgs in wells.items():
        for block,gener in bgs:
            if (block,gener) in grouping:
                if well != grouping[(block,gener)]:
                    err = 'Spec %s a gener (%s,%s) is associated ' \
                          'with diff wells: %s != %s' % (fspec,
                          block, gener, grouping[(block,gener)], well)
                    raise Exception(err)
            grouping[(block,gener)] = well
    return grouping

def default_grouping(bgs, custom_grouping=None):
    """ given a list of (block, gener), automatically returns grouping, which is
        a dict {(block, gener): 'GRP NAME', ...}

    custom_grouping can be specified, eg.
        {
            "dry1": ["WK234", "WK238", "WK240", ...],
            "dry2": ["WK236", "WK216", "WK118", ...],
            ...
        }
    """
    specified = {}
    if custom_grouping:
        for gg, bb in custom_grouping.items():
            for b in bb:
                if b in specified:
                    err = 'Custom grouping has (%s) in two groups: %s and %s' % (
                        b, specified[b], gg)
                    raise Exception(err)
                specified[b] = gg

    delimiter = '-'
    group_by_first_n_chars = 2
    grouping = {}
    for bg in bgs:
        use_key = bg
        if isinstance(bg, tuple):
            use_key = bg[1]
        if bg in specified:
            grp = specified[bg]
        elif '-' in use_key:
            grp = use_key.split('-')[0]
        else:
            grp = use_key[:group_by_first_n_chars]
        grouping[bg] = grp
    return grouping

def export_mass_enth_flat_csv(fname, mass, enth, times,
                              to_mass_unit=assign_kgs,
                              to_enth_unit=assign_jkg,
                              to_time_unit=assign_sec):
    # _nu is in new unit
    times_nu, time_lbl = to_time_unit(times)
    mass_lbl = to_mass_unit(1.0)[1]
    enth_lbl = to_enth_unit(1.0)[1]
    with open(fname, 'w') as f:
        f.write(f"Well Name,Year,Date (DD/MM/YYYY),Rate ({mass_lbl}),Enthalpy ({enth_lbl})\n")
        for name in sorted(mass.keys()):
            if len(mass[name]) == 0:
                continue
            mass_nu = [to_mass_unit(x)[0] for x in mass[name]]
            enth_nu = [to_enth_unit(x)[0] for x in enth[name]]
            for tt,mm,ee in zip(times_nu, mass_nu, enth_nu):
                dd = year_fraction_to_date_str(tt)
                f.write(f"{name},{tt},{dd},{mm},{ee}\n")
        f.write("\n")

def plot_num_wells_used(pdf_pages, times,
                        well_mass, spec_json,
                        title_prefix='',
                        to_time_unit=assign_sec,
                        silent=True):
    wellstacks = wellstacks_from_spec(spec_json)
    times_nu, time_lbl = to_time_unit(times)
    stacks = {}
    for stackname, wells in wellstacks.items():
        num_wells_used, i_first_used = wells_required_over_time(well_mass, wells)
        if not silent:
            print(f"    - stack: {stackname}")
        for w in wells:
            i_used = i_first_used[w]
            if not silent:
                if i_used is not None:
                    print(f"       {w} {time_lbl} {times_nu[i_used]}")
                else:
                    print(f"       {w} -")

        fig, ax = plt.subplots(figsize=(8.5,13))
        ax.step(times_nu, num_wells_used, where='post', marker='o', color='black')
        jumps = np.diff(num_wells_used, prepend=num_wells_used[0])
        for i_time, jump in enumerate(jumps):
            if jump > 0:
                # wells first used at this time
                newly_opened = [w for w in wells if i_time == i_first_used[w]]
                if newly_opened:
                    label = "\n".join(newly_opened)
                    y = num_wells_used[i_time]
                    ax.text(times_nu[i_time], y + 0.1, label, ha='center', va='bottom', fontsize=9,
                            bbox=dict(boxstyle='round,pad=0.3', fc='white', ec='black'))

        ax.set_xlabel("Time")
        ax.set_ylabel("Number of wells")
        ax.set_title(f"{title_prefix} [{stackname}] with {len(wells)} wells")
        ax.set_ylim(bottom=0)
        ymin, ymax = ax.get_ylim()
        ax.set_ylim(ymin, ymax * 1.2)  # or ymax + 0.5
        pdf_pages.savefig(fig)
        plt.close(fig)
        fig.clear()
        stacks[stackname] = {
            'wells': wells,
            'num_wells_used': num_wells_used,
            'first_used': i_first_used,
        }
    return stacks

def plot_mass_enth_groups(pdf_pages, times,
                          grouped_mass, grouped_enth, grouped_tags,
                          title_prefix='',
                          to_mass_unit=assign_kgs,
                          to_enth_unit=assign_jkg,
                          to_time_unit=assign_sec,):

    MAX_N_LEGENDS = 50
    # _nu is in new unit
    times_nu, time_lbl = to_time_unit(times)

    return_mass, return_enth = {}, {} # in original unit
    for grp in sorted(grouped_mass.keys()):
        if len(grouped_mass[grp]) == 0:
            continue
        total_mass, avg_enth = average_enthalpy(grouped_mass[grp], grouped_enth[grp])
        return_mass[grp] = total_mass
        return_enth[grp] = avg_enth

        all_mass_nu = [to_mass_unit(m)[0] for m in grouped_mass[grp][::-1]]
        all_enth_nu = [to_enth_unit(e)[0] for e in grouped_enth[grp][::-1]]
        avg_enth, enth_lbl = to_enth_unit(avg_enth)
        total_mass, mass_lbl = to_mass_unit(total_mass)

        fig, (ax1, ax2) = plt.subplots(2,1, sharex=True)
        # ax3.get_shared_x_axes().join(ax2, ax3)
        fig.set_figwidth(8.5)
        fig.set_figheight(13)

        ax1.set_title('%s Group %s' % (title_prefix, grp))
        # ax1.axis([0, 10, 0, 10])
        # ax1.text(5, 5, '\n'.join(grouped_tags[grp]), ha='center', va='center', wrap=True)
        # ax1.tick_params(left=False, right=False, bottom=False,
        #                 labelleft=False, labelbottom = False)

        ax1.stackplot(times_nu, np.array(all_mass_nu), labels=grouped_tags[grp][::-1])
        box = ax1.get_position()
        ax1.set_position([box.x0, box.y0, box.width * 0.8, box.height])
        handles, labels = ax1.get_legend_handles_labels()
        leg = ax1.legend(handles[::-1][:MAX_N_LEGENDS+1],
                         labels[::-1][:MAX_N_LEGENDS]+['...'],
                         loc="center left",
                         bbox_to_anchor=(1, -0.1)) # loc='upper left'
        # leg = ax1.legend(handles[::-1], labels[::-1]) # loc='upper left'
        # leg.draggable()
        # reverse=True, requires matplotlib >= 3.7
        ax1.set_ylabel('Massflow (%s)' % mass_lbl)
        ax1.grid(True)

        for gen,enth in zip(grouped_tags[grp], all_enth_nu):
            ax2.plot(times_nu, enth)
        if len(grouped_tags[grp]) > 1:
            ax2.plot(times_nu, avg_enth, 'ko:', label='Average')
            ax2.legend()
        try:
            # new lim based on nonzeros
            all_enth_nu = np.array(all_enth_nu)
            all_enth_nz = all_enth_nu[np.nonzero(all_enth_nu)]
            minval = np.min(all_enth_nz)
            maxval = np.max(all_enth_nz)
            margin = (maxval-minval) * 0.10
            ax2.set_ylim(minval-margin, maxval+margin)
            # ax2.margins(y=0.1)
        except ValueError:
            # means no non-zero values at all?
            pass
        box = ax2.get_position()
        ax2.set_position([box.x0, box.y0, box.width * 0.8, box.height])
        ax2.set_ylabel('Enthalpy (%s)' % enth_lbl)
        ax2.set_xlabel('Time (%s)' % time_lbl)
        ax2.grid(True)

        # plt.show()
        pdf_pages.savefig(fig)
        plt.close(fig)
        fig.clear()
    return return_mass, return_enth

def group_mass_enth(grouping, mass, enth, tag=str, custom_sort_key=sorted):
    """ returns grouped mass/enth/names by following grouping, a dict with the
        same keys of mass and enth.

        tag is a function that creates a string from each name.  custom_sort_key
        is a func that can be specified to order how tags appear within each
        group.
    """
    if isinstance(custom_sort_key, list):
        def order():
            for n in custom_sort_key:
                yield n
    else:
        def order():
            for n in custom_sort_key(mass.keys()):
                yield n
    grouped_mass, grouped_enth = {}, {}
    grouped_tags = {}
    for name in order():
        # grp = grouping[name]
        grp = grouping.get(name, None)
        if grp is None:
            print("Skipping (%s, %s), does not have group." % name)
        if grp not in grouped_mass:
            grouped_mass[grp] = []
            grouped_enth[grp] = []
            grouped_tags[grp] = []
        grouped_mass[grp].append(mass[name])
        grouped_enth[grp].append(enth[name])
        grouped_tags[grp].append(tag(name))
    return grouped_mass, grouped_enth, grouped_tags


def ever_non_zero(history):
    """ return array of 0s and 1s after the history has been active (ever != 0)

    eg. history = [0, 0, 2.5, 0, 0] gives [0, 0, 1, 1, 1]

    This is useful for counting number of geners/wells required.
    """
    if isinstance(history, list):
        history = np.array(history)
    return np.cumsum(history != 0.0).clip(max=1)

def wells_required_over_time(flowrates, welllist=None):
    """ compute number of wells required over time.

    returns a tuple of :
        single array of total wells ever flowed
        a dict of index when each well is first flowed, None if never flowed

    flowrates : dict of named flowrates, each is a numpy array
    welllist : optional if given, only count these wells
    """
    ever_flowed, first_flowed = {}, {}
    for name, flow in flowrates.items():
        if welllist is not None:
            if name not in welllist:
                continue
        ever_flowed[name] = ever_non_zero(flow)
        first_flowed[name] = np.argmax(ever_flowed[name])
        if ever_flowed[name][first_flowed[name]] == 0:
            first_flowed[name] = None
    return sum(ever_flowed.values()), first_flowed

class TestWellsRequired:
    def test(self):
        expected = [
            ([0, 0, 0.01, 0, 0],      [0, 0, 1, 1, 1]),
            ([0, 0, -0.01, 0.5, 3],   [0, 0, 1, 1, 1]),
            ([0, 0, 0.01, 0.5, 0,],   [0, 0, 1, 1, 1]),
            ([0.1, 0, 0.01, 0.5, 0,], [1, 1, 1, 1, 1]),
            ([0., 0, 0.0, 0., 0,],    [0, 0, 0, 0, 0]),
        ]
        for inp, out in expected:
            res = ever_non_zero(inp)
            assert list(res) == out

        rates_dict = {f"w{i}": flow for i, (flow, ever) in enumerate(expected)}
        wells_needed, first_used = wells_required_over_time(rates_dict)
        assert list(wells_needed) == [1, 1, 4, 4, 4]
        assert first_used == {
            'w0': 2,
            'w1': 2,
            'w2': 2,
            'w3': 0,
            'w4': None, # if never flowed, return None
        }

        wells_needed, first_used = wells_required_over_time(rates_dict, welllist=["w0", "w3"])
        assert list(wells_needed) == [1, 1, 2, 2, 2]


def main():
    with open('settings.json', 'r') as f:
        cfg = json.load(f)

    match_block, match_gener = cfg['production']['block'], cfg['production']['gener']
    rebn, regn = re.compile(match_block), re.compile(match_gener)
    use_spec = False
    if 'use_spec_json' in cfg:
        use_spec = cfg['use_spec_json']
    custom_grouping = cfg.get('custom_grouping', {})

    MASSNAME, ENTHNAME = 'Generation rate', 'Enthalpy'

    for sdir in cfg['dir_to_extract']:
        spec_filename = os.path.join(sdir, 'scenario_spec.json')
        if use_spec and os.path.isfile(spec_filename):
            with open(spec_filename, 'r') as f:
                spec_json = json.load(f)
            flsts = [os.path.join(sdir, sim['filename']+'.h5') for sim in spec_json['simulations']]
            fdats = [os.path.join(sdir, sim['filename']+'.dat') for sim in spec_json['simulations']]
            offset_year = spec_json['date_offset']
            print('Found optional spec %s, plot spec based grouping.' % spec_filename)
            use_spec = True
            output_postfix = '_spec'
        else:
            flsts = sorted(glob.glob('%s/wai*.h5' % sdir))
            fdats = [fn.replace('.h5', '.dat') for fn in flsts]
            offset_year = 0.0
            use_spec = False
            output_postfix = '_gengrp'

        # use basename (the last part of the the normalized path) as sname
        sname = os.path.basename(os.path.normpath(sdir))
        sname = external_name(cfg, sname)
        print("Extracting scenario '%s' from directory: %s" % (sname, sdir))

        fname_pdf = 'figs_prd_%s%s.pdf' % (sname, output_postfix)
        fname_csv1 = 'figs_prd_%s_wells.csv' % (sname)
        fname_csv2 = 'figs_prd_%s_groups.csv' % (sname)
        fname_cache = '_extract_prd_grps_%s%s.pkl' % (sname, output_postfix)


        use_cache = False
        if os.path.exists(fname_cache):
            ans = input('Cache file for %s exists. Use it? (y/n) ' % sname)
            if ans.lower().startswith('y'):
                use_cache = True

        if use_cache:
            print('Loading cached data from %s' % fname_cache)
            with open(fname_cache, 'rb') as f:
                gener_hist, bns_gns = pickle.load(f)
        else:
            print('Extracting from listings for %s...' % sname)
            lsts = [t2listingh5(fn) for fn in flsts]
            bns_gns = matching_geners_from_lsts(lsts, rebn, regn)
            if bns_gns:
                gener_hist, _ = get_gener_history(
                    [b for b,g in bns_gns], [g for b,g in bns_gns],
                    lsts, cols=[MASSNAME, ENTHNAME], silent=True)
            else:
                gener_hist, bns_gns = {}, []

            if not gener_hist:
                print("No generators found for %s. Skipping." % sname)
                continue

            print('Saving cache to %s' % fname_cache)
            with open(fname_cache, 'wb') as f:
                pickle.dump((gener_hist, bns_gns), f)

        print('    Extracted/Loaded %d generators for %s.' % (len(bns_gns), sname))

        b, g = bns_gns[0]
        times = gener_hist[(b, g, MASSNAME)][0]

        grp_total_mass = {(b,g): gener_hist[(b, g, MASSNAME)][1] for b,g in bns_gns}
        grp_total_enth = {(b,g): gener_hist[(b, g, ENTHNAME)][1] for b,g in bns_gns}

        # each set of plots requires the following four settings
        # grouping, tag, sort, mass unit

        pdf_pages = PdfPages(fname_pdf)

        # spec based grouping, only if able to load spec json file
        if use_spec:
            orig_mass, orig_enth = grp_total_mass, grp_total_enth
            grouping = grouping_from_spec(spec_json)
            from pprint import pprint
            print('<<<<< use spec set >>>>>')
            # pprint(grouping)
            def gener2tag(gen_key):
                return '(%s,%s)' % gen_key
            grouped_mass, grouped_enth, grouped_tags = group_mass_enth(
                grouping, orig_mass, orig_enth,
                tag=gener2tag, # cleaner of gener name: (bname,gname)
                custom_sort_key=bns_gns, # follow order of original list
                )
            grp_total_mass, grp_total_enth = plot_mass_enth_groups(
                pdf_pages, times,
                grouped_mass, grouped_enth, grouped_tags,
                title_prefix=sname,
                to_mass_unit=to_thr_rev,
                to_enth_unit=to_kjkg,
                to_time_unit=partial(to_year, start_year=offset_year),
                )
            print('    Spec groups DONE.')

            print('<<<<< number of wells used >>>>>')
            stacks = plot_num_wells_used(
                pdf_pages, times,
                grp_total_mass,
                spec_json,
                title_prefix=sname,
                to_time_unit=partial(to_year, start_year=offset_year),
                silent=False)
            print('    number of wells used DONE.')


        print(f'    Exporting flat csv: {fname_csv1} ... ', end='')
        export_mass_enth_flat_csv(fname_csv1, grp_total_mass, grp_total_enth, times,
            to_mass_unit=to_thr_rev,
            to_enth_unit=to_kjkg,
            to_time_unit=partial(to_year, start_year=offset_year),
            )
        print('DONE.')

        # standard set of plots
        orig_mass, orig_enth = grp_total_mass, grp_total_enth
        grouping = default_grouping(list(grp_total_mass.keys()), custom_grouping=custom_grouping)
        print('<<<<< standard set >>>>>')
        # pprint(grouping)
        grouped_mass, grouped_enth, grouped_tags = group_mass_enth(
            grouping, orig_mass, orig_enth,
            tag=str,
            custom_sort_key=sorted,
            )
        grp_total_mass, grp_total_enth = plot_mass_enth_groups(
            pdf_pages, times,
            grouped_mass, grouped_enth, grouped_tags,
            title_prefix=sname,
            to_mass_unit=to_ktd_rev,
            to_enth_unit=to_kjkg,
            to_time_unit=partial(to_year, start_year=offset_year),
            )
        print('    Standard groups DONE.')

        print(f'    Exporting flat csv: {fname_csv2} ... ', end='')
        export_mass_enth_flat_csv(fname_csv2, grp_total_mass, grp_total_enth, times,
            to_mass_unit=to_thr_rev,
            to_enth_unit=to_kjkg,
            to_time_unit=partial(to_year, start_year=offset_year),
            )
        print('DONE.')

        # final single group contains everything
        orig_mass, orig_enth = grp_total_mass, grp_total_enth
        grouping = {k: "All Geners: '%s'" % match_gener for k in grp_total_mass.keys()}
        print('<<<<< all >>>>>')
        # pprint(grouping)
        grouped_mass, grouped_enth, grouped_tags = group_mass_enth(
            grouping, orig_mass, orig_enth,
            tag=str,
            custom_sort_key=sorted,
            )
        grp_total_mass, grp_total_enth = plot_mass_enth_groups(
            pdf_pages, times,
            grouped_mass, grouped_enth, grouped_tags,
            title_prefix=sname,
            to_mass_unit=to_ktd_rev,
            to_enth_unit=to_kjkg,
            to_time_unit=partial(to_year, start_year=offset_year),
            )
        print('    Final group DONE.')

        pdf_pages.close()

    return 0

if __name__ == '__main__':
    exit(main())




from t2listing import *
from t2data import *

from gimfu.t2listingh5 import t2listingh5

from gimfu.multiple_listings_plotting import get_block_history
from gimfu.multiple_listings_plotting import get_gener_history
from gimfu.scenario_extraction import external_name

import re
import os
import glob
import json
from copy import deepcopy
from pprint import pprint as pp
import pickle
import numpy as np

import xlwt
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages


def normalname_match(name_wc, name):
    """ return true if the name_wc (wild cast * supported) match name """
    is_match = True
    for i in range(5):
        if name_wc[i] != '*':
            if name_wc[i] != name[i]:
                is_match = False
                break
    return is_match

def get_gener_curve(bname,gname,dat):
    for g in dat.generatorlist:
        if g.name == gname and g.block == bname:
            if abs(g.ltab) > 1:
                enth = g.time
                pcut = g.rate
                return enth,pcut

def to_kjkg(xs):
    return [x/1000. for x in xs]

def to_bar(xs):
    return [x/1.e5 for x in xs]

def to_year(xs,start_year=0.0):
    sec_in_year = 60. * 60. * 24. * 365.25
    return [x / sec_in_year + start_year for x in xs]

def to_tday(xs):
    return [x * 86.4 for x in xs]

def to_tday_rev(xs):
    return [-x * 86.4 for x in xs]

def to_thr(xs):
    return [x * 3.6 for x in xs]

def to_thr_rev(xs):
    return [-x * 3.6 for x in xs]

def to_MJday_rev(xs):
    # production heatflow -J/s to MJ/day
    return [-x * 60.*60.*24./1.0e6 for x in xs]

def to_tdaybar(xs):
    # deliverability terms (mobility-density-PI)
    # kg/s/Pa to t/day/bar
    return [x * 60.*60.*24.*1.e5/1.e3 for x in xs]


def gener_pdf(pdf_pages, gener_hist, block_hist, datgen_hist, notes=''):
    """ plot history of gener on deliverability into a pdf file

    If notes is a string then it will be printed on all pages, if notes is a
    dict with (block, gener) as key, then each page can have its own notes.

    - each figure/page is a single generator
    - each page contains sub figures:
        + Pcutoff curve, wellbore pressu vs. enthalpy
        + notes
        + reservoir pressure history and Pcutoff
        + enthalpy history
        + massflow history
        + steamflow history
    """
    raise NotImplementedError

def plot_prd_geners(pdf_fname, bns_gns, gener_hist, block_hist, datgen_hist, curves, notes=""):
    FIGURE_DIR = 'non_standatd_figures'
    if not os.path.exists(FIGURE_DIR):
        os.makedirs(FIGURE_DIR)
    pdf_pages = PdfPages(pdf_fname)
    for bn,gn in bns_gns:
        curve_x, curve_y = curves[(bn,gn)]

        ts = gener_hist[(bn, gn, 'Enthalpy')][0]
        gx = gener_hist[(bn, gn, 'Enthalpy')][1]
        gy = block_hist[(bn, 'Pressure')][1]
        mass = gener_hist[(bn, gn, 'Generation rate')][1]
        steam = gener_hist[(bn, gn, 'Steam sepa.')][1]
        cap = datgen_hist[(bn, gn, 'Mass Cap')]
        reqps = gener_hist[(bn, gn, 'Wellbore pressure')][1]


        curve_x, curve_y = to_kjkg(curve_x), to_bar(curve_y)
        gx, gy = to_kjkg(gx), to_bar(gy)
        reqps = to_bar(reqps)
        ts = to_year(ts,1953.0)
        mass, steam = to_thr_rev(mass), to_thr(steam)
        cap = to_thr(cap)

        fig = plt.figure(figsize=(16.0,10.0))

        fig.add_subplot(321)
        plt.plot(gx,gy,'.-')
        plt.plot(curve_x, curve_y,'.-')
        plt.plot(gx[0],gy[0],'ro')
        plt.xlabel('Enthalpy (kJ/kg)')
        plt.ylabel('Pressure (bara)')
        plt.ylim(0.0,220.0)
        plt.xlim(750.0,3000.0)
        plt.grid(True)
        plt.legend(labels=['Actual','Delv. curve'],loc='upper right')
        plt.title('Pcutoff Curve (%s,%s)' % (bn,gn))

        fig.add_subplot(322)
        plt.axis([0, 10, 0, 10])
        plt.text(5, 5, notes, ha='center', va='center', wrap=True)

        fig.add_subplot(323)
        plt.plot(ts, gy,'.-')
        plt.plot(ts, reqps,'.-')
        plt.plot(ts[0],gy[0],'ro')
        plt.xlabel('Year')
        plt.ylabel('Pressure (bar)')
        plt.legend(labels=['Actual','Pcutoff'],loc='upper right')
        plt.grid(True)

        fig.add_subplot(324)
        plt.plot(ts, gx,'.-')
        plt.plot(ts[0],gx[0],'ro')
        plt.xlabel('Year')
        plt.ylabel('Enthalpy (kJ/kg)')
        plt.grid(True)

        fig.add_subplot(325)
        plt.plot(ts,mass,'.-')
        plt.plot(ts,cap, '.--')
        plt.plot(ts[0],mass[0],'ro')
        plt.xlabel('Year')
        plt.ylabel('Mass Flow Rate (t/hr)')
        plt.grid(True)

        fig.add_subplot(326)
        plt.plot(ts,steam,'.-')
        plt.plot(ts[0],steam[0],'ro')
        plt.xlabel('Year')
        plt.ylabel('Steam Flow Rate (t/hr)')
        plt.grid(True)

        # plt.savefig(FIGURE_DIR+'/'+gn+'.png')
        #orientation='portrait', papertype='a4'
        pdf_pages.savefig(fig)
        plt.close(fig)
    pdf_pages.close()

def export_prd_geners(xls_fname, bns_gns, gener_hist, block_hist, datgen_hist, curves):
    if not bns_gns:
        print('export_prd_geners(): No matched geners found, do nothing.')
        return
    wb = xlwt.Workbook()
    ws = wb.add_sheet('production blocks')
    # plot one by one
    for ii,(bn,gn) in enumerate(bns_gns):
        # time, massflow, enthalpy, pressure, steamflow
        ts = gener_hist[(bn, gn, 'Enthalpy')][0]
        fs = gener_hist[(bn, gn, 'Generation rate')][1]
        es = gener_hist[(bn, gn, 'Enthalpy')][1]
        ps = block_hist[(bn, 'Pressure')][1]
        ss = gener_hist[(bn, gn, 'Steam sepa.')][1]
        pws = gener_hist[(bn, gn, 'Wellbore pressure')][1]
        dvs = gener_hist[(bn, gn, 'Deliverability')][1]
        cs = datgen_hist[(bn, gn, 'Mass Cap')]


        # calc heatflow kg/s * J/kg = J/s -> * 60*60*24/1e6 = MJ/day
        hs = [f*e for f,e in zip(fs,es)]
        hs = to_MJday_rev(hs)
        ts = to_year(ts,1953.0)
        fs = to_tday_rev(fs)
        cs = to_tday(cs)
        es = to_kjkg(es)
        ps = to_bar(ps)
        ss = to_tday(ss)
        pws = to_bar(pws)
        dvs = to_tdaybar(dvs)

        ofsx = ii * 8 # X, ii * number of variables
        ofsy = 0   # Y
        # TIMES
        # print(ofsx, ofsy, bn)
        ws.write(ofsx, ofsy, bn)
        ws.write(ofsx, ofsy+1, gn)
        ws.write(ofsx, ofsy+2, 'Times (year)')
        for jj,v in enumerate(ts):
            ws.write(ofsx, 3+jj, v)
        # MASS
        ofsx += 1 # X
        ws.write(ofsx, ofsy, bn)
        ws.write(ofsx, ofsy+1, gn)
        ws.write(ofsx, ofsy+2, 'Massflow (t/day)')
        for jj,v in enumerate(fs):
            ws.write(ofsx, 3+jj, v)
        # ENTH
        ofsx += 1 # X
        ws.write(ofsx, ofsy, bn)
        ws.write(ofsx, ofsy+1, gn)
        ws.write(ofsx, ofsy+2, 'Enthalpy (kJ/kg)')
        for jj,v in enumerate(es):
            ws.write(ofsx, 3+jj, v)
        # HEAT
        ofsx += 1 # X
        ws.write(ofsx, ofsy, bn)
        ws.write(ofsx, ofsy+1, gn)
        ws.write(ofsx, ofsy+2, 'Heatflow (MJ/day)')
        for jj,v in enumerate(hs):
            ws.write(ofsx, 3+jj, v)
        # PRESSURE
        ofsx += 1 # X
        ws.write(ofsx, ofsy, bn)
        ws.write(ofsx, ofsy+1, gn)
        ws.write(ofsx, ofsy+2, 'Pressure (bar)')
        for jj,v in enumerate(ps):
            ws.write(ofsx, 3+jj, v)
        # WELLBORE PRESSURE
        ofsx += 1 # X
        ws.write(ofsx, ofsy, bn)
        ws.write(ofsx, ofsy+1, gn)
        ws.write(ofsx, ofsy+2, 'Wellbore Pressure (bar)')
        for jj,v in enumerate(pws):
            ws.write(ofsx, 3+jj, v)
        # DELIVERABILITY
        ofsx += 1 # X
        ws.write(ofsx, ofsy, bn)
        ws.write(ofsx, ofsy+1, gn)
        ws.write(ofsx, ofsy+2, 'Deliverability (t/day/bar)')
        for jj,v in enumerate(dvs):
            ws.write(ofsx, 3+jj, v)
        # MASS CAP
        ofsx += 1 # X
        ws.write(ofsx, ofsy, bn)
        ws.write(ofsx, ofsy+1, gn)
        ws.write(ofsx, ofsy+2, 'Massflow Cap (t/day)')
        for jj,v in enumerate(cs):
            ws.write(ofsx, 3+jj, v)
    # save exported xls
    wb.save(xls_fname)


def export_prd_geners_flat(xls_fname, bns_gns, gener_hist, block_hist, datgen_hist):
    if not bns_gns:
        print('export_prd_geners_flat(): No matched geners found, do nothing.')
        return
    # flat xls
    wb = xlwt.Workbook()
    ws = wb.add_sheet('production blocks')

    ofsx = 0
    ofsy = 0   # Y
    ws.write(ofsx, ofsy, 'Times (year)'); ofsy += 1
    ws.write(ofsx, ofsy, 'Block'); ofsy += 1
    ws.write(ofsx, ofsy, 'Gener'); ofsy += 1
    ws.write(ofsx, ofsy, 'Massflow (t/day)'); ofsy += 1
    ws.write(ofsx, ofsy, 'Enthalpy (kJ/kg)'); ofsy += 1
    ws.write(ofsx, ofsy, 'Heatflow (MJ/day)'); ofsy += 1
    ws.write(ofsx, ofsy, 'Pressure (bar)'); ofsy += 1
    ws.write(ofsx, ofsy, 'Wellbore Pressure (bar)'); ofsy += 1
    ws.write(ofsx, ofsy, 'Deliverability (t/day/bar)'); ofsy += 1
    ws.write(ofsx, ofsy, 'Massflow Cap (t/day)'); ofsy += 1

    # plot one by one
    for ii,(bn,gn) in enumerate(bns_gns):
        # time, massflow, enthalpy, pressure, steamflow
        ts = gener_hist[(bn, gn, 'Enthalpy')][0]
        fs = gener_hist[(bn, gn, 'Generation rate')][1]
        es = gener_hist[(bn, gn, 'Enthalpy')][1]
        ps = block_hist[(bn, 'Pressure')][1]
        ss = gener_hist[(bn, gn, 'Steam sepa.')][1]
        pws = gener_hist[(bn, gn, 'Wellbore pressure')][1]
        dvs = gener_hist[(bn, gn, 'Deliverability')][1]
        cs = datgen_hist[(bn, gn, 'Mass Cap')]

        # calc heatflow kg/s * J/kg = J/s -> * 60*60*24/1e6 = MJ/day
        hs = [f*e for f,e in zip(fs,es)]
        hs = to_MJday_rev(hs)
        ts = to_year(ts,1953.0)
        fs = to_tday_rev(fs)
        cs = to_tday(cs)
        es = to_kjkg(es)
        ps = to_bar(ps)
        ss = to_tday(ss)
        pws = to_bar(pws)
        dvs = to_tdaybar(dvs)
        for time, mass, enth, heat, pres, pwb, delv, cap in zip(ts, fs, es, hs, ps, pws, dvs, cs):
            ofsx += 1
            ofsy = 0   # Y
            ws.write(ofsx, ofsy, time); ofsy += 1
            ws.write(ofsx, ofsy, bn); ofsy += 1
            ws.write(ofsx, ofsy, gn); ofsy += 1
            ws.write(ofsx, ofsy, mass); ofsy += 1
            ws.write(ofsx, ofsy, enth); ofsy += 1
            ws.write(ofsx, ofsy, heat); ofsy += 1
            ws.write(ofsx, ofsy, pres); ofsy += 1
            ws.write(ofsx, ofsy, pwb); ofsy += 1
            ws.write(ofsx, ofsy, delv); ofsy += 1
            ws.write(ofsx, ofsy, cap); ofsy += 1
    # save exported xls
    wb.save(xls_fname)


def main_(sdir, sname, block_name, gener_name):
    MAKE_PLOTS = True
    MAKE_XLS = True
    MAKE_XLS_FLAT = True

    flsts = sorted(glob.glob('%s/wai*.h5' % sdir))
    fdats = [fn.replace('.h5', '.dat') for fn in flsts]
    out_fname = 'prd_%s' % sname
    cache_fname = '_extract_prd_geners_%s.pkl' % sname
    use_cache = False
    if os.path.exists(cache_fname):
        ans = input('Cache file %s exists. Use it? (y/n) ' % cache_fname)
        if ans.lower() == 'y':
            use_cache = True

    if use_cache:
        print(f'Loading from cache {cache_fname} ...')
        with open(cache_fname, 'rb') as f:
            bns_gns, gener_hist, block_hist, datgen_hist, curves, lst_fulltimes = pickle.load(f)
    else:
        print(f'Loading {len(flsts)} model results...')
        lsts = [t2listingh5(fn) for fn in flsts]
        print(f"Loading {len(fdats)} dat files...")
        dats = [t2data(fn) for fn in fdats]

        # fulltimes for each lst
        lst_fulltimes = [lst.fulltimes for lst in lsts]

        # find all matching generators
        # collect Pcutoff curves
        rebn, regn = re.compile(block_name), re.compile(gener_name)
        bns_gns = []
        curves = {}
        for dat in dats:
            print('  in dat file %s' % dat.filename)
            for g in dat.generatorlist:
                if re.match(rebn, g.block) and re.match(regn, g.name):
                    if (g.block,g.name) not in bns_gns:
                        bns_gns.append((g.block,g.name))
                        if abs(g.ltab) > 1:
                            # g.time is enthalpy, g.rate is Pcutoff
                            curves[(g.block,g.name)] = (g.time, g.rate)

        variables = ['Generation rate', 'Enthalpy', 'Steam sepa.', 'Wellbore pressure', 'Deliverability']
        gener_hist, gener_bgcns = get_gener_history([b for b,g in bns_gns],
                                                    [g for b,g in bns_gns],
                                                    lsts, cols=variables,
                                                    silent=False)

        blocks = list(set([b for b,g,cn in gener_bgcns]))
        block_hist, block_bcns = get_block_history(blocks, lsts, ['Pressure'])

        # compute the *fake* history of g.hg
        def get_gen_data(gen=None, dname=None):
            d = {
                'Mass Cap': 'hg',
            }
            if dname is None:
                return d.keys()
            else:
                return getattr(gen, d[dname])
        g_dnames = get_gen_data()
        datgen_hist = {}
        for dat,ts in zip(dats,lst_fulltimes):
            vs = {}
            for bn,gn in bns_gns:
                for cn in g_dnames:
                    value = 0.0
                    if (bn,gn) in dat.generator:
                        gen = dat.generator[(bn,gn)]
                        if gen.type in ['DMAT']:
                            value = get_gen_data(gen, cn)
                    if (bn,gn,cn) not in datgen_hist:
                        datgen_hist[(bn,gn,cn)] = np.array([], dtype=np.float64)
                    datgen_hist[(bn,gn,cn)] = np.concatenate((datgen_hist[(bn,gn,cn)],
                                                              np.ones_like(ts) * value))

        with open(cache_fname, 'wb') as f:
            print(f"Writing/overwriting cache file {cache_fname} ...")
            pickle.dump([bns_gns, gener_hist, block_hist, datgen_hist, curves, lst_fulltimes], f)


    if MAKE_PLOTS:
        print(f"Plotting ...")
        plot_prd_geners('figs_%s.pdf' % out_fname,
            bns_gns, gener_hist, block_hist, datgen_hist, curves,
            notes="\n".join(flsts))

    if MAKE_XLS:
        print(f"Exporting ...")
        export_prd_geners('%s.xls' % out_fname,
            bns_gns, gener_hist, block_hist, datgen_hist, curves)

    if MAKE_XLS_FLAT:
        print(f"Exporting (flat) ...")
        export_prd_geners_flat('%s_flat.xls' % out_fname,
            bns_gns, gener_hist, block_hist, datgen_hist)

def main():
    with open('settings.json', 'r') as f:
        cfg = json.load(f)

    bname, gname = cfg['production']['block'], cfg['production']['gener']
    for sdir in cfg['dir_to_extract']:
        # use basename (the last part of the the normalized path) as sname
        sname = os.path.basename(os.path.normpath(sdir))
        sname = external_name(cfg, sname)
        print("Extracting scenario '%s' from directory: %s" % (sname, sdir))

        main_(sdir, sname, bname, gname)

    print("Done.")
    return 0

if __name__ == '__main__':
    exit(main())




from t2listing import *
from t2data import *

from gimfu.t2listingh5 import t2listingh5

from gimfu.multiple_listings_plotting import get_block_history
from gimfu.multiple_listings_plotting import get_gener_history
from gimfu.scenario_extraction import external_name

import os
import glob
import re
import json
from copy import deepcopy

import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages

import xlwt


def normalname_match(name_wc, name):
    """ return true if the name_wc (wild cast * supported) match name """
    is_match = True
    for i in xrange(5):
        if name_wc[i] != '*':
            if name_wc[i] != name[i]:
                is_match = False
                break
    return is_match


def to_kjkg(xs):
    return [x/1000. for x in xs]

def to_bar(xs):
    return [x/1.e5 for x in xs]

def to_year(xs,start_year=0.0):
    sec_in_year = 60. * 60. * 24. * 365.25
    return [x / sec_in_year + start_year for x in xs]

def to_tday(xs):
    return [x * 86.4 for x in xs]

def to_tday_rev(xs):
    return [-x * 86.4 for x in xs]

def plot_inj_geners(sdir, scenario_name, block_name, gener_name):
    FIGURE_DIR = 'non_standatd_figures'
    if not os.path.exists(FIGURE_DIR):
        os.makedirs(FIGURE_DIR)

    sname = scenario_name
    out_fname = 'inj_%s' % sname.replace('/', '_').replace('\\', '_')
    lst_fnames = sorted(glob.glob('%s/wai*.h5' % sdir))
    dat_fnames = [fn.replace('.h5', '.dat') for fn in lst_fnames]

    lsts, dats = [], []
    print('found listing files:')
    for flst,fdat in zip(lst_fnames, dat_fnames):
        print('    loading %s, %s ...' % (flst, fdat))
        lsts.append(t2listingh5(flst))
        dats.append(t2data(fdat))

    bname, gname = block_name, gener_name
    rebn, regn = re.compile(bname), re.compile(gname)
    columns =  ['Generation rate']
    units =    ['t/day']
    converts = [to_tday]
    tbl,bgcs = get_gener_history(bname, gname, lsts, cols=columns)

    # collect gener info from t2data, and fake it into time series as from lst
    # collect place holder first: all matching gners from all dats
    imak_pcutoff = {}
    for dat in dats:
        for g in dat.generatorlist:
            if re.match(rebn, g.block) and re.match(regn, g.name) and g.type == 'IMAK':
                gid = (g.block, g.name)
                if gid not in imak_pcutoff:
                        imak_pcutoff[gid] = []
    for dat, lst in zip(dats, lsts):
        pcutoff = {}
        for i,g in enumerate(dat.generatorlist):
            if re.match(rebn, g.block) and re.match(regn, g.name) and g.type == 'IMAK':
                pcutoff[(g.block, g.name)] = g.hg
        for gid in imak_pcutoff:
            # use value if gener in this dat, otherwise use 0.0
            if gid in pcutoff:
                imak_pcutoff[gid] += [pcutoff[gid]] * lst.num_fulltimes
            else:
                imak_pcutoff[gid] += [0.0] * lst.num_fulltimes


    wb = xlwt.Workbook()
    ws = wb.add_sheet('injection well blocks')
    for ii,(b,g,c) in enumerate(bgcs):
        # TIMES
        ofsx = ii * 2 # X
        ofsy = 0      # Y
        ws.write(ofsx, ofsy, b)
        ws.write(ofsx, ofsy+1, g)
        ws.write(ofsx, ofsy+2, 'Times (year)')
        values = to_year(tbl[(b,g,c)][0], start_year=1953.0)
        for jj,v in enumerate(values):
            ws.write(ofsx, 3+jj, v)
        for cn,u,fc in zip(columns, units, converts):
            # MASS etc
            ofsx += 1 # X
            ws.write(ofsx, ofsy, b)
            ws.write(ofsx, ofsy+1, g)
            ws.write(ofsx, ofsy+2, '%s (%s)' % (cn,u))
            values = fc(tbl[(b,g,c)][1])
            for jj,v in enumerate(values):
                ws.write(ofsx, 3+jj, v)
    wb.save(out_fname + '.xls')

    # get pressure for the blocks
    blocklist = list(set([b for b,g,c in bgcs]))
    columns = ['Pressure']
    ptbl,bcs = get_block_history(blocklist, lsts, cols=columns)

    pdf_pages = PdfPages('figs_%s.pdf' % out_fname)

    bgs = []
    for b,g,c in bgcs:
        if (b,g) not in bgs:
            bgs.append((b,g))

    for b,g in bgs:
        fig = plt.figure(figsize=(16.0,10.0))

        fig.add_subplot(221)
        xs = to_year(tbl[(b,g,'Generation rate')][0], start_year=1953.0)
        ys = to_tday(tbl[(b,g,'Generation rate')][1])
        plt.plot(xs, ys)
        plt.xlabel('Year')
        plt.ylabel('Mass Flow Rate (t/day)')
        plt.title('Generator (%s,%s)' % (b,g))
        plt.grid(True)

        fig.add_subplot(222)
        xs = to_year(ptbl[b, 'Pressure'][0], start_year=1953.0)
        ys = to_bar(ptbl[b, 'Pressure'][1])
        plt.plot(xs, ys)
        cutoff = to_bar(imak_pcutoff[(b,g)])
        plt.plot(xs, cutoff)
        plt.xlabel('Year')
        plt.ylabel('Pressure (bar)')
        plt.title('Block %s' % b)
        plt.grid(True)

        fig.add_subplot(223)
        plt.axis([0, 10, 0, 10])
        ss = '\n'.join([s for s in lst_fnames])
        plt.text(5, 5, ss, ha='center', va='center', wrap=True)
        plt.grid(False)

        pdf_pages.savefig(fig)
        plt.close(fig)

    pdf_pages.close()

def main():
    with open('settings.json', 'r') as f:
        cfg = json.load(f)

    bname, gname = cfg['injection']['block'], cfg['injection']['gener']
    for sdir in cfg['dir_to_extract']:
        # use basename (the last part of the the normalized path) as sname
        sname = os.path.basename(os.path.normpath(sdir))
        sname = external_name(cfg, sname)
        print("Extracting scenario '%s' from directory: %s" % (sname, sdir))

        plot_inj_geners(sdir, sname, bname, gname)

    return 0

if __name__ == '__main__':
    exit(main())




import csv

from mulgrids import well

class WellPlan:
    """
    Loads and parses a planned well file exported from Seequent's Leapfrog.
    The file is expected to have two sections:
    1. A key-value header section.
    2. A tabular data section for the well track.
    The sections are separated by two empty lines.
    """
    def __init__(self, filename):
        """
        Initializes and loads the well data from the specified file.

        Args:
            filename: The path to the .csv file.
        """
        self.filename = filename
        self.header = {}
        self.data = {}
        self._load_file()

    def _load_file(self):
        """
        Internal method to handle the file parsing logic.
        """
        with open(self.filename, 'r', encoding='utf-8') as f:
            lines = f.readlines()

        # Find the index of the separator (two consecutive '""' lines)
        separator_index = -1
        for i, line in enumerate(lines):
            if line.strip() == '""':
                if i + 1 < len(lines) and lines[i+1].strip() == '""':
                    separator_index = i
                    break
        
        if separator_index == -1:
            raise ValueError(f"Could not find the '""' separator in file: {self.filename}")

        header_lines = lines[:separator_index]
        data_lines = lines[separator_index + 2:]

        # --- Parse Header Section ---
        for line in header_lines:
            line = line.strip()
            if not line:
                continue
            
            parts = line.split(',', 1)
            if len(parts) == 2:
                key, value = parts
                # Attempt to convert numerical values to float
                try:
                    self.header[key] = float(value)
                except ValueError:
                    # Handle special case for Location coordinates
                    if key == 'Location':
                        self.header[key] = [float(coord) for coord in value.split(',')]
                    else:
                        self.header[key] = value

        # --- Parse Data Section ---
        if not data_lines:
            return

        # Use the csv module to handle potential quoting in headers or data
        reader = csv.reader(data_lines)
        
        try:
            column_headers = next(reader)
        except StopIteration:
            return # No data rows found

        self.data = {header: [] for header in column_headers}
        
        for row in reader:
            if not row: # Skip empty lines
                continue
            for i, value in enumerate(row):
                if i < len(column_headers):
                    header = column_headers[i]
                    try:
                        # Convert all data values to float
                        self.data[header].append(float(value))
                    except (ValueError, IndexError):
                        # If conversion fails or row is malformed, append as is
                        self.data[header].append(value)


    def __repr__(self):
        well_id = self.header.get('Well ID', 'N/A')
        return f"PlannedWell(filename='{self.filename}', well_id='{well_id}')"


    def block_at_layer(self, geo, layer, cache_qtree=True):
        """ return the feed block for the planned well at the specified layer

        TODO instead of simple approach of pos at the layer centre, should
             consider well curverture and trace through the layer thickness
             see:
                 https://github.com/pro-well-plan/well_profile
        """
        # get cached qtree if already generated
        geo_qtree = None
        if cache_qtree:
            if not hasattr(geo, '_qtree'):
                geo._qtree = geo.column_quadtree()
            geo_qtree = geo._qtree
        # use mulgrid/well to handle the position and block lookup
        data_columns = ['Easting', 'Northing', 'Elevation']
        pos = list(zip(*[self.data[c] for c in data_columns]))
        well_name = ''
        geo_well = well(well_name, pos)
        feed_pos = geo_well.elevation_pos(geo.layer[layer].centre)
        feed_block = geo.block_name_containing_point(feed_pos, qtree=geo_qtree)
        return feed_block

    def add_to_mulgrid(self, geo, name=None):
        """ add a well object into mulgrid """
        data_columns = ['Easting', 'Northing', 'Elevation']
        pos = list(zip(*[self.data[c] for c in data_columns]))
        if name is None:
            # default anme uses Well ID, trim to last 5 characters
            well_name = self.header.get('Well ID', 'WELL1')[-5:]
        else:
            well_name = name
        w = well(well_name, pos)
        geo.add_well(w)






""" shared functions that deals with extracting scenario results
"""

def external_name(cfg, sname):
    if isinstance(cfg, str):
        with open(cfg, 'r') as f:
            import json
            cfg = json.load(f)
    ext_names = cfg['external_scenario_names']
    if '-' in sname:
        parts = sname.split('-')
        new_parts = []
        for s in parts:
            try:
                s = ext_names[s]
            except KeyError:
                pass
            new_parts.append(s)
        return '-'.join(new_parts)
    else:
        return ext_names[sname]


def matching_geners_from_lsts(lsts, block_name, gener_name):
    """ return a list of (block, gener) that match the block_name and gener_name
    """
    import re
    rebn, regn = block_name, gener_name
    if isinstance(block_name, str):
         rebn = re.compile(block_name)
    if isinstance(gener_name, str):
         regn = re.compile(gener_name)
    bns_gns = []
    for lst in lsts:
        for b,g in lst.generation.row_name:
            if re.match(rebn, b) and re.match(regn, g):
                if (b, g) not in bns_gns:
                    bns_gns.append((b, g))
    return bns_gns


def average_enthalpy(massrates, enthalpies):
    """ returns total massrates and the averaged enthalpies

    massrates should be a list of numpy arrays, enthalpies should be a list of
    numpy arrays

    (lengths of the lists should be the same, length/shape of all numpy arrays
    should be the same)
    """
    import numpy as np
    totalheat = np.zeros_like(massrates[0])
    totalmass = np.zeros_like(massrates[0])
    for mass,enth in zip(massrates, enthalpies):
        totalmass += mass
        totalheat += mass * enth
    avgenths = np.divide(totalheat, totalmass, out=np.zeros_like(totalheat),
                         where=totalmass!=0)
    return totalmass, avgenths


def to_year(xs, start_year=0.0):
    sec_in_year = 60. * 60. * 24. * 365.25
    return xs / sec_in_year + start_year, 'Year'

def to_kjkg(xs):
    return xs/1000., 'kJ/kg'

def to_bar(xs):
    return xs/1.e5, 'bar'

def to_thr(xs):
    return xs * 3.6, 't/hr'

def to_thr_rev(xs):
    return xs * -3.6, 't/hr'

def to_ktd(xs):
    return xs * 0.0864, 'ktd'

def to_ktd_rev(xs):
    return xs * -0.0864, 'ktd'

def assign_kgs(xs):
    """ no convert """
    return xs, 'kg/s'

def assign_jkg(xs):
    """ no convert """
    return xs, 'J/kg'

def assign_sec(xs):
    """ no convert """
    return xs, 'sec'

